from .batchers import FixedBatchSizeMapper
from .collators import (
    FromTokenizerListCollatorMapper,
    FromTokenizerTensorCollatorMapper,
    ListCollatorMapper,
    TensorCollatorMapper,
)
from .converters import Python2TorchMapper, Torch2PythonMapper
from .fields import ChangeFieldsMapper, MakeFieldMapper
from .loaders import (
    CsvLoaderMapper,
    HuggingFaceDatasetLoaderMapper,
    JsonlLoaderMapper,
)
from .multiseq import (
    AttentionMaskSequencePaddingMapper,
    LabelsMaskerMapper,
    MakeAttentionMaskMapper,
    MultiSequenceStriderMapper,
    SequencesConcatenateMapper,
    SingleValueToSequenceMapper,
    TokensSequencesPaddingMapper,
    TokenTypeIdsSequencePaddingMapper,
)
from .shape import FlattenMapper, UnpackingMapper
from .text import FtfyMapper
from .tokenize import PaddingMapper, TokenizerMapper, ValidUnicodeMapper
from .types import BinarizerMapper, CastMapper, LookupMapper, OneHotMapper

__all__ = [
    "AttentionMaskSequencePaddingMapper",
    "BinarizerMapper",
    "CastMapper",
    "ChangeFieldsMapper",
    "CsvLoaderMapper",
    "FixedBatchSizeMapper",
    "FlattenMapper",
    "FromTokenizerListCollatorMapper",
    "FromTokenizerTensorCollatorMapper",
    "FtfyMapper",
    "HuggingFaceDatasetLoaderMapper",
    "JsonlLoaderMapper",
    "LabelsMaskerMapper",
    "ListCollatorMapper",
    "LookupMapper",
    "MakeAttentionMaskMapper",
    "MakeFieldMapper",
    "MultiSequenceStriderMapper",
    "OneHotMapper",
    "PaddingMapper",
    "Python2TorchMapper",
    "SequencesConcatenateMapper",
    "SingleValueToSequenceMapper",
    "TensorCollatorMapper",
    "TokenizerMapper",
    "TokensSequencesPaddingMapper",
    "TokenTypeIdsSequencePaddingMapper",
    "Torch2PythonMapper",
    "UnpackingMapper",
    "ValidUnicodeMapper",
]



from typing import TYPE_CHECKING, Any, Dict, Optional, TypeVar, Union

from necessary import necessary
from trouting import trouting

from ..base.mappers import SingleBaseMapper
from ..base.types import TransformElementType

with necessary("datasets", soft=True) as HUGGINGFACE_DATASET_AVAILABLE:
    if HUGGINGFACE_DATASET_AVAILABLE or TYPE_CHECKING:
        from datasets.arrow_dataset import Dataset
        from datasets.iterable_dataset import IterableDataset

        HuggingFaceDataset = TypeVar(
            "HuggingFaceDataset", Dataset, IterableDataset
        )
        from datasets.features import features


HF_CAST_DICT = {
    int: "int64",
    float: "float32",
    bool: "bool",
    str: "string",
}


class CastMapper(SingleBaseMapper):
    """Casts one or more fields in a dataset to a new type."""

    def __init__(self, cast_map: Dict[str, type]):
        """Initializes a CastMapper.

        Args:
            cast_map (Dict[str, type]): A mapping from field names to the
                type to cast them to. In case a dataset is a HuggingFace
                dataset, the type will be converted to the appropriate value
                for the datasets.Value class.
        """
        self.cast_map = cast_map
        super().__init__(
            input_fields=list(cast_map.keys()),
            output_fields=list(cast_map.keys()),
        )

    @trouting
    def map(  # type: ignore
        self,
        dataset: Any,
        **map_kwargs: Any,
    ) -> Any:
        # we need this map to be able to add the new interface below
        # and handle types for which we don't have a new interface but our
        # parent class has one
        return super().map(dataset, **map_kwargs)

    def _cast(self, value: Any, type_: Optional[type] = None) -> Any:
        if type_ is None:
            return value
        elif isinstance(value, list):
            return [self._cast(value=v, type_=type_) for v in value]
        elif isinstance(value, dict):
            return {
                k: self._cast(value=v, type_=type_) for k, v in value.items()
            }
        else:
            return type_(value)

    def transform(self, data: TransformElementType) -> TransformElementType:
        return {
            k: self._cast(value=v, type_=self.cast_map.get(k, None))
            for k, v in data.items()
        }

    if HUGGINGFACE_DATASET_AVAILABLE:

        def _build_feature_definition(
            self,
            def_n: Union[
                dict, features.ClassLabel, features.Sequence, features.Value
            ],
            type_: type,
        ) -> Union[features.ClassLabel, features.Sequence, features.Value]:
            """A helper function to build a new feature definition in case
            the dataset is a HuggingFace dataset.

            Args:
                def_n (Union[ features.ClassLabel, features.Sequence, features.
                    Value ]): The current feature definition from
                    datasets.features dictionary.
                type_ (type): The type to cast the feature to.

            Returns:
                Union[ features.ClassLabel, features.Sequence,
                    features.Value ]: The new feature definition.
            """

            # TODO[soldni]: document better!

            if (
                t_str := HF_CAST_DICT.get(type_, None)  # pyright: ignore
            ) is None:
                raise ValueError(
                    f"Unsupported type {type_} for HuggingFace Dataset"
                )

            if isinstance(def_n, features.Sequence):
                if isinstance(def_n.feature, dict):
                    new_feature = {
                        k: self._build_feature_definition(v, type_)
                        for k, v in def_n.feature.items()
                    }
                else:
                    new_feature = self._build_feature_definition(
                        def_n.feature, type_
                    )

                new_definition = features.Sequence(feature=new_feature)

            elif isinstance(def_n, features.ClassLabel):
                new_names = list(set(type_([n for n in def_n.names])))
                new_definition = features.ClassLabel(
                    names=new_names,
                    num_classes=len(new_names),
                )

            elif isinstance(def_n, features.Value):
                new_definition = features.Value(dtype=t_str)
            else:
                raise ValueError(f"Unsupported feature definition {def_n}")

            return new_definition

        @map.add_interface(dataset=(Dataset, IterableDataset))
        def map_huggingface_dataset(
            self,
            dataset: HuggingFaceDataset,
            **map_kwargs: Any,
        ) -> HuggingFaceDataset:
            dataset = super().map(dataset, **map_kwargs)
            for field_name, field_type in self.cast_map.items():
                dataset = dataset.cast_column(
                    field_name,
                    self._build_feature_definition(
                        def_n=dataset.features[field_name], type_=field_type
                    ),
                )
            return dataset


class BinarizerMapper(CastMapper):
    """Binarizes a field in a dataset."""

    def __init__(self, field: str, threshold: float) -> None:
        """Initializes a BinarizerMapper.

        Args:
            field (str): The field to binarize.
            threshold (float): The threshold to use for binarization.

        """
        super().__init__(cast_map={field: int})
        self.threshold = threshold

    def transform(self, data: TransformElementType) -> TransformElementType:
        field_name, *_ = self.input_fields

        if isinstance(data[field_name], list):
            return {
                field_name: [
                    1 if v > self.threshold else 0 for v in data[field_name]
                ]
            }
        else:
            return {field_name: 1 if data[field_name] > self.threshold else 0}


class LookupMapper(CastMapper):
    def __init__(self, field_name: str, lookup_table: Dict[Any, Any]):

        source_types = set([type(k) for k in lookup_table.keys()])
        target_types = set([type(v) for v in lookup_table.values()])

        if len(source_types) > 1:
            raise ValueError(
                "Lookup source values must be of the same type, "
                f"but got {source_types}"
            )
        if len(target_types) > 1:
            raise ValueError(
                "Lookup target values must be of the same type, "
                f"but got {target_types}"
            )

        super().__init__(cast_map={field_name: list(target_types)[0]})
        self.field_name = field_name
        self.lookup_table = lookup_table

    def transform(self, data: TransformElementType) -> TransformElementType:
        return {
            **data,
            self.field_name: self.lookup_table[data[self.field_name]],
        }


class OneHotMapper(CastMapper):
    """One-hot encodes a field in a dataset."""

    def __init__(self, num_classes: int, field_name: str) -> None:
        """Initializes a OneHotMapper.

        Args:
            num_classes (int): The number of classes.
            field (str): The field to one-hot encode.

        """
        super().__init__(cast_map={field_name: int})
        self.field_name = field_name
        self.num_classes = num_classes

    def transform(self, data: TransformElementType) -> TransformElementType:
        return {
            **data,
            self.field_name: [
                1 if i == data[self.field_name] else 0
                for i in range(self.num_classes)
            ],
        }



from functools import reduce
from itertools import chain
from typing import Any, Tuple, Type, Union

from .mappers.abstract import AbstractBaseMapper


class Pipeline:
    """A pipeline is a sequence of mappers that are applied to a dataset.

    Pipelines can be created by chaining two mappers together or by chaining
    a pipeline with another pipeline or mapper. Use operators << and >> to
    chain mappers or pipelines, or call the "chain" method on the class itself.

    To execute a pipeline, call the pipeline with a dataset as the first
    argument; any additional arguments are passed to the map method of each
    mapper in the pipeline.
    """

    mappers: Tuple[AbstractBaseMapper, ...]

    def __init__(self, *mappers: AbstractBaseMapper) -> None:
        self.mappers = mappers

    def __repr__(self: "Pipeline") -> str:
        mappers_it = (repr(m) for m in self.mappers)
        return f'Pipeline({" -> ".join(mappers_it)})'

    def __str__(self: "Pipeline") -> str:
        mappers_it = (str(m) for m in self.mappers)
        return f'Pipeline({" -> ".join(mappers_it)})'

    def __lshift__(
        self: "Pipeline", other: Union[AbstractBaseMapper, "Pipeline"]
    ) -> "Pipeline":
        return self.chain(other, self)

    def __rshift__(
        self: "Pipeline", other: Union[AbstractBaseMapper, "Pipeline"]
    ) -> "Pipeline":
        return self.chain(self, other)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Pipeline):
            return False
        if len(self.mappers) != len(other.mappers):
            return False
        for my_mapper, theirs_mapper in zip(self.mappers, other.mappers):
            if my_mapper != theirs_mapper:
                return False
        return True

    @classmethod
    def chain(
        cls: Type["Pipeline"],
        *mappers_or_pipelines: Union[AbstractBaseMapper, "Pipeline"],
    ) -> "Pipeline":
        """Create a new pipeline by chaining two mappers/pipelines together."""

        def _to_pip(
            mapper_or_pipeline: Union[AbstractBaseMapper, "Pipeline"]
        ) -> "Pipeline":
            if isinstance(mapper_or_pipeline, AbstractBaseMapper):
                mapper_or_pipeline = Pipeline(mapper_or_pipeline)
            return mapper_or_pipeline

        return Pipeline(
            *chain.from_iterable(
                _to_pip(m_or_p).mappers for m_or_p in mappers_or_pipelines
            )
        )

    def map(self: "Pipeline", dataset: Any, **map_kwargs: Any) -> Any:
        """Transform a dataset by applying this pipeline's mappers."""

        # IMPLEMENTATION FOR DEBUG PURPOSES
        # for mapper in self.mappers:
        #     dataset = mapper.map(dataset, **map_kwargs)
        #     if not dataset:
        #         breakpoint()
        # return dataset

        def _map(dataset: Any, mapper: AbstractBaseMapper) -> Any:
            return mapper.map(dataset, **map_kwargs)

        return reduce(_map, self.mappers, dataset)



from abc import ABCMeta, abstractmethod
from typing import Any, Generic, Iterable, List, TypeVar

from ..types import TransformElementType  # type: ignore

D = TypeVar("D")
S = TypeVar("S")


class AbstractBaseMapper(Generic[D, S], metaclass=ABCMeta):
    """An abstract implementation of a Mapper"""

    __slots__: List[str] = ["input_fields", "output_fields"]
    input_fields: List[str]
    output_fields: List[str]
    ...

    @abstractmethod
    def map(self, dataset: D, **map_kwargs: Any) -> D:
        """Transform a dataset by applying this mapper's transform method.

        Args:
            dataset (DatasetType): The dataset to transform.
            **map_kwargs: Additional keyword arguments to pass to the
                transform method. By default, this is empty.

        Returns:
            dataset (DatasetType): The transformed dataset.
        """
        ...

    @abstractmethod
    def transform(self, data: S) -> S:
        """Transform a single sample of a dataset. This method should be
        overridden by actual mapper implementations.

        Args:
            self (BaseMapper): _description_
            data (Any): _description_

        Raises:
            NotImplementedError: _description_

        Returns:
            Any: _description_
        """
        ...


class AbstractSingleBaseMapper(AbstractBaseMapper):
    """An abstract implementation of a Mapper that operates on a single
    element."""

    @abstractmethod
    def transform(self, data: TransformElementType) -> TransformElementType:
        """Transform a single sample of a dataset. This method should be
        overridden by actual mapper implementations.

        Args:
            data (TransformElementType): The sample to transform. This is
                a single sample dictionary with string keys and values of
                any type.

        Returns:
            TransformElementType: The transformed sample. This is a single
                sample dictionary with string keys and values of any type.
                The keys can be different from the input keys.
        """
        ...


class AbstractBatchedBaseMapper(AbstractBaseMapper, metaclass=ABCMeta):
    """An abstract implementation of a Mapper that operates on a batch of
    elements."""

    @abstractmethod
    def transform(
        self, data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:
        """Transform a batch of data. This method should be overridden by
        actual mapper implementations.

        Args:
            data (Iterable[TransformElementType]): The batch of data to
                transform; it is an iterable of dictionaries, where each
                dictionary is a sample (i.e. a Dict[str, Any]).

        Returns:
            Iterable[TransformElementType]: Iterable of transformed data.
                Each element in the iterable is a dictionary with str keys
                and values of any type. The number of elements in the
                iterable may be different from the number of samples in the
                input.
        """
        ...



from collections import abc
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from necessary import necessary
from trouting import trouting

from ..types import TransformBatchType, TransformElementType
from .abstract import (
    AbstractBaseMapper,
    AbstractBatchedBaseMapper,
    AbstractSingleBaseMapper,
)

with necessary("datasets", soft=True) as HUGGINGFACE_DATASET_AVAILABLE:
    if HUGGINGFACE_DATASET_AVAILABLE or TYPE_CHECKING:
        from datasets.arrow_dataset import Dataset
        from datasets.iterable_dataset import IterableDataset

        HuggingFaceDataset = TypeVar(
            "HuggingFaceDataset", Dataset, IterableDataset
        )


class MapMethodInterfaceMixIn(AbstractBaseMapper):
    def _check_fields_datasets(
        self,
        provided_fields: Union[Iterable[str], None],
        expected_fields: Sequence[str],
    ) -> None:
        if provided_fields is None:
            return

        provided_fields_set = set(provided_fields)

        for field in expected_fields:
            if field not in provided_fields_set:
                raise ValueError(f"Field {field} not found in dataset")

    def _get_iterator_and_column_names_list_dataset(
        self,
        dataset: Sequence[TransformElementType],
    ) -> Tuple[Iterable[TransformElementType], Set[str]]:
        """Given an iterable dataset, return the name of the columns
        as well as an iterator over the dataset."""

        dataset_iter = iter(dataset)
        try:
            first_element = next(dataset_iter)
        except StopIteration:
            return iter([]), set()

        column_names: Set[str] = {str(e) for e in first_element.keys()}

        dataset_iter_chained: Iterable[TransformElementType] = chain(
            (first_element,), dataset_iter
        )
        return dataset_iter_chained, column_names

    def _batch_transform_huggingface_datasets(
        self, data: TransformBatchType
    ) -> TransformBatchType:
        """Unrolls a datasets.Dataset batch, which is a dictionary of
        <features, list of feature values for each sample> into a iterable
        of dictionaries that can be passed to the transform function."""

        keys = [k for k in data.keys()]

        # _index_fn ensures that, between when we unpack
        # the sequence of samples in TrasformBatchType, and when we
        # pack them into a list of dictionaries, we always get the
        # same order of features. This is important because we don't
        # want one feature value accidentally getting mapped to the
        # wrong feature name
        def _index_fn(t: Tuple[str, Any]) -> int:
            k, _ = t
            return keys.index(k)

        to_transform_iterable = (
            dict(zip(keys, sample))
            for sample in zip(
                *(v for _, v in sorted(data.items(), key=_index_fn))
            )
        )
        transformed_batch: Dict[str, List[Any]] = {}
        for transformed_sample in self.transform(to_transform_iterable):
            for k, v in transformed_sample.items():
                transformed_batch.setdefault(k, []).append(v)

        return transformed_batch

    @trouting
    def map(  # type: ignore
        self,
        dataset: Any,
        **map_kwargs: Any,
    ) -> Any:
        raise ValueError(
            f"I don't know how to a dataset of type {type(dataset)}"
        )

    @map.add_interface(dataset=list)
    def map_list_of_dicts(
        self,
        dataset: Sequence[TransformElementType],
        **map_kwargs: Any,
    ) -> Sequence[TransformElementType]:
        """Transform a dataset by applying this mapper's transform method.

        Args:
            dataset (DatasetType): The dataset to transform.
            remove_columns (Optional[bool], optional): If True, remove discard
                any columns that are in the input dataset, but are not returned
                by the transform method. Defaults to False.
            map_kwargs (Any, optional): Additional keyword arguments to pass to
                the transform method. By default, this is empty; other
                implementations may use this.
        """

        # explicitly casting to a boolean since this is all that is
        # supported by the simple mapper.
        # TODO[lucas]: maybe support specifying which fields to keep?
        remove_columns = bool(map_kwargs.pop("remove_columns", False))

        if isinstance(dataset, abc.Sequence):
            self._check_fields_datasets(
                provided_fields=dataset[0].keys(),
                expected_fields=self.input_fields,
            )

        if isinstance(self, AbstractBatchedBaseMapper):
            (
                dataset_it,
                columns_names,
            ) = self._get_iterator_and_column_names_list_dataset(dataset)
            transformed_dataset_it = self.transform(dataset_it)

            if remove_columns:
                transformed_dataset_it = (
                    {k: v for k, v in elem.items() if k in columns_names}
                    for elem in transformed_dataset_it
                )
            transformed_dataset = list(transformed_dataset_it)

        elif isinstance(self, AbstractSingleBaseMapper):
            if remove_columns:
                # we don't care about the original columns
                transformed_dataset = [
                    self.transform(sample) for sample in dataset
                ]
            else:
                # user wants to keep the columns, so we merge the new fields
                # with the old fields, while keeping the new ones if there
                # is a name conflict
                transformed_dataset = [
                    {**sample, **self.transform(sample)} for sample in dataset
                ]
        else:
            raise TypeError(
                "Mapper must inherit a SingleBaseMapper or a BatchedBaseMapper"
            )

        if isinstance(dataset, abc.Sequence):
            self._check_fields_datasets(
                provided_fields=transformed_dataset[0].keys(),
                expected_fields=self.output_fields,
            )

        return transformed_dataset

    if HUGGINGFACE_DATASET_AVAILABLE:

        @map.add_interface(dataset=(Dataset, IterableDataset))
        def map_huggingface_dataset(
            self,
            dataset: HuggingFaceDataset,
            **map_kwargs: Any,
        ) -> HuggingFaceDataset:
            self._check_fields_datasets(
                provided_fields=dataset.features.keys(),
                expected_fields=self.input_fields,
            )

            if isinstance(self, AbstractBatchedBaseMapper):
                transformed_dataset = dataset.map(
                    self._batch_transform_huggingface_datasets,
                    **{**map_kwargs, "batched": True},  # type: ignore
                )
            elif isinstance(self, AbstractSingleBaseMapper):
                transformed_dataset = dataset.map(self.transform, **map_kwargs)
            else:
                raise TypeError(
                    "Mapper but be either a SingleBaseMapper or "
                    "a BatchedBaseMapper"
                )

            self._check_fields_datasets(
                provided_fields=transformed_dataset.features.keys(),
                expected_fields=self.output_fields,
            )

            return transformed_dataset



from ..mappers import *  # noqa
from ..utils import SmashedWarnings

SmashedWarnings.deprecation(
    "smashed.interfaces.simple is deprecated; "
    "import from smashed.mappers instead."
)


class Dataset(list):
    def __new__(cls, *args, **kwargs) -> list:  # type: ignore
        SmashedWarnings.deprecation(
            "smashed.interfaces.simple.Dataset is deprecated; "
            "simply use a list of dictionaries with str keys instead."
        )
        return list(*args, **kwargs)



import json
from collections import abc
from csv import DictReader
from typing import TYPE_CHECKING, Any, Iterable, List, Optional, TypeVar, Union

from necessary import necessary
from trouting import trouting

from ..base.mappers import BatchedBaseMapper
from ..base.types import TransformElementType

with necessary("datasets", soft=True) as HUGGINGFACE_DATASET_AVAILABLE:
    if HUGGINGFACE_DATASET_AVAILABLE or TYPE_CHECKING:
        from datasets.arrow_dataset import Dataset
        from datasets.dataset_dict import DatasetDict, IterableDatasetDict
        from datasets.iterable_dataset import IterableDataset
        from datasets.load import load_dataset

        HuggingFaceDataset = TypeVar(
            "HuggingFaceDataset", Dataset, IterableDataset
        )

with necessary("smart_open", soft=True) as SMART_OPEN_AVAILABLE:
    if SMART_OPEN_AVAILABLE:
        from smart_open import open


class HuggingFaceDatasetLoaderMapper(BatchedBaseMapper):
    def __init__(
        self,
        **load_datasets_args: Any,
    ):
        if "path" not in load_datasets_args:
            raise ValueError(
                "datasets.load_dataset requires a path to a dataset."
            )

        self.load_datasets_args = load_datasets_args

    @trouting
    def map(  # type: ignore
        self,
        dataset: Any,
        **map_kwargs: Any,
    ) -> Any:
        # we need this map to be able to add the new interface below
        # and handle types for which we don't have a new interface but our
        # parent class has one
        super().map(dataset, **map_kwargs)

    @map.add_interface(dataset=type(None))
    def map_none(  # type: ignore
        self,
        dataset: None,
        **map_kwargs: Any,
    ) -> Any:
        # this loader
        return self.transform([])

    if HUGGINGFACE_DATASET_AVAILABLE:

        def transform(
            self, data: Iterable[TransformElementType]
        ) -> Union[Dataset, IterableDataset]:
            dataset = load_dataset(**self.load_datasets_args)

            if isinstance(dataset, (DatasetDict, IterableDatasetDict)):
                raise ValueError(
                    "HuggingFaceDatasetLoader only supports loading a single "
                    f"dataset, but the provided dataset is a {type(dataset)}. "
                    "Please provide a `split` to this mapper."
                )

            return dataset


class CsvLoaderMapper(BatchedBaseMapper):
    def __init__(
        self,
        paths_field: str,
        headers: Optional[List[str]] = None,
        encoding: str = "utf-8",
        **dict_reader_args: Any,
    ) -> None:
        self.paths_field = paths_field
        self.headers = headers
        self.encoding = encoding
        self.dict_reader_args = dict_reader_args
        super().__init__(input_fields=[paths_field], output_fields=headers)

    def transform(
        self, data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:

        for row in data:
            paths = row[self.paths_field]
            if not isinstance(paths, abc.Sequence) or isinstance(paths, str):
                paths = [paths]

            for path in paths:
                with open(path, mode="rt", encoding=self.encoding) as f:
                    if self.headers:
                        yield from DictReader(
                            f, fieldnames=self.headers, **self.dict_reader_args
                        )
                    else:
                        yield from DictReader(f, **self.dict_reader_args)


class JsonlLoaderMapper(BatchedBaseMapper):
    def __init__(
        self,
        paths_field: str,
        encoding: str = "utf-8",
    ) -> None:
        self.paths_field = paths_field
        self.encoding = encoding
        super().__init__(input_fields=[paths_field])

    def transform(
        self, data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:

        for row in data:
            paths = row[self.paths_field]
            if not isinstance(paths, abc.Sequence) or isinstance(paths, str):
                paths = [paths]

            for path in paths:
                with open(path, mode="rt", encoding=self.encoding) as f:
                    for ln in f:
                        yield json.loads(ln)



from ...base.mappers import SingleBaseMapper
from ...base.types import TransformElementType


class OneVsOtherAnnotatorMapper(SingleBaseMapper):
    def __init__(
        self,
        input_field: str = "annotations",
        label_field: str = "labels",
        preds_field: str = "preds",
        position: int = 0,
    ) -> None:
        super().__init__(
            input_fields=[input_field],
            output_fields=[label_field, preds_field],
        )
        self.position = position

    def transform(self, data: TransformElementType) -> TransformElementType:
        input_field_name, *_ = self.input_fields
        label_field_name, preds_field_name, *_ = self.output_fields

        if len(data[input_field_name]) < 2:
            raise ValueError(
                f"Expected at least 2 annotations, "
                f"got {len(data[input_field_name])}"
            )

        other_labels = [
            v
            for i, v in enumerate(data[input_field_name])
            if i != self.position
        ]

        return {
            preds_field_name: data[input_field_name][self.position],
            label_field_name: sum(other_labels) / len(other_labels),
        }



from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Iterable, List, Optional, Union

from ..types import TransformElementType
from .abstract import (
    AbstractBaseMapper,
    AbstractBatchedBaseMapper,
    AbstractSingleBaseMapper,
)
from .interfaces import MapMethodInterfaceMixIn

if TYPE_CHECKING:
    from ..pipeline import Pipeline


__all__ = ["SingleBaseMapper", "BatchedBaseMapper"]


class LshiftRshiftMixIn(AbstractBaseMapper):
    def __init__(
        self,
        input_fields: Optional[List[str]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> None:
        """Create a new Mapper.

        Args:
            input_fields (Optional[List[str]], optional): The fields expected
                by this mapper. If None is provided, the mapper will not
                check for the presence of any input fields. Defaults to None.
            output_fields (Optional[List[str]], optional): The fields produced
                by this mapper after transformation. If None is provided, the
                mapper will not validate the output of transform. Defaults to
                None.
        """
        self.input_fields = input_fields or []
        self.output_fields = output_fields or []

    def __lshift__(
        self,
        other: Union["LshiftRshiftMixIn", "Pipeline"],
    ) -> "Pipeline":
        """Create a new Pipeline by combining this mapper with another."""
        # avoid circular import
        from ..pipeline import Pipeline

        return Pipeline(self) << other

    def __rshift__(
        self,
        other: Union["LshiftRshiftMixIn", "Pipeline"],
    ) -> "Pipeline":
        """Create a new Pipeline by combining this mapper with another."""
        return other << self


class SingleBaseMapper(
    MapMethodInterfaceMixIn,
    LshiftRshiftMixIn,
    AbstractSingleBaseMapper,
    metaclass=ABCMeta,
):
    """An abstract implementation of a Mapper that operates on a single
    element. All mappers that operate on a single element should subclass
    this class.

    Actual mapper implementations should override the transform method.
    The transform method should accept a single sample dictionary as input
    and return a single sample dictionary as output.
    """

    @abstractmethod
    def transform(self, data: TransformElementType) -> TransformElementType:
        """Transform a single sample of a dataset. This method should be
        overridden by actual mapper implementations.

        Args:
            data (TransformElementType): The sample to transform. This is
                a single sample dictionary with string keys and values of
                any type.

        Returns:
            TransformElementType: The transformed sample. This is a single
                sample dictionary with string keys and values of any type.
                The keys can be different from the input keys.
        """
        raise NotImplementedError("Mapper subclass must implement transform")


class BatchedBaseMapper(
    MapMethodInterfaceMixIn,
    LshiftRshiftMixIn,
    AbstractBatchedBaseMapper,
    metaclass=ABCMeta,
):
    """An abstract implementation of a Mapper that operates on a batch of
    elements. All mappers that operate on a batch should subclass this
    class.

    Actual mapper implementations should override the transform method.
    The transform method should accept a iterator of dictionaries as input,
    and return a iterator of dictionaries as output. The number of samples
    returned may be different from the number of samples in the input.
    """

    @abstractmethod
    def transform(
        self, data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:
        """Transform a batch of data. This method should be overridden by
        actual mapper implementations.

        Args:
            data (Iterable[TransformElementType]): The batch of data to
                transform; it is an iterable of dictionaries, where each
                dictionary is a sample (i.e. a Dict[str, Any]).

        Returns:
            Iterable[TransformElementType]: Iterable of transformed data.
                Each element in the iterable is a dictionary with str keys
                and values of any type. The number of elements in the
                iterable may be different from the number of samples in the
                input.
        """
        raise NotImplementedError("Mapper subclass must implement transform")



from itertools import chain
from typing import Any, Dict, Iterable, List, Optional

from ..base.mappers import BatchedBaseMapper, SingleBaseMapper
from ..base.types import TransformElementType


class FlattenMapper(SingleBaseMapper):
    def __init__(self, field: str) -> None:
        super().__init__(input_fields=[field], output_fields=[field])

    def transform(self, data: TransformElementType) -> TransformElementType:
        field_name, *_ = self.input_fields

        flattened_field = data[field_name]

        if len(flattened_field) > 0:
            while isinstance(flattened_field[0], list):
                flattened_field = list(chain.from_iterable(flattened_field))

        return {field_name: flattened_field}


class UnpackingMapper(BatchedBaseMapper):
    _DRP_EXTRA = "drop"
    _RPT_EXTRA = "repeat"

    def __init__(
        self,
        fields_to_unpack: Optional[List[str]] = None,
        fields_to_ignore: Optional[List[str]] = None,
        ignored_behavior: Optional[str] = None,
    ) -> None:
        """Unpacks some or all fields in a dataset sample.
        Useful when features contain a list of values, but you want to
        want individual values for each sample.

        Args:
            fields_to_unpack (Optional[List[str]], optional): List of fields to
                unpack. When not None, the other fields will be repeated (if
                `ignore_behavior` is not set to "repeat") or dropped (if
                `ignore_behavior` is set to "drop"). Only one between
                `fields_to_unpack` and `fields_to_ignore` can be set. Defaults
                to None.
            fields_to_ignore (Optional[List[str]], optional): List of fields to
                ignore. When not None, the fields provided will be dropped/
                duplicated, while the others will be unpacked. Only one between
                `fields_to_unpack` and `fields_to_ignore` can be set. Defaults
                to None.
            ignore_behavior (Optional[str]b, optional): How to handle fields
                that are not unpacked. Can be "drop" or "repeat". Defaults to
                None. Must be set when either `fields_to_unpack` or
                `fields_to_ignore` is not None.
        """

        if fields_to_unpack is not None and fields_to_ignore is not None:
            raise ValueError(
                "Must specify only one of `fields_to_unpack` "
                "or `fields_to_ignore`"
            )

        if (
            fields_to_unpack is not None or fields_to_ignore is not None
        ) and ignored_behavior not in {self._DRP_EXTRA, self._RPT_EXTRA}:
            raise ValueError(
                f"When specifying `fields_to_unpack` or `fields_to_ignore`, "
                f"`ignore_behavior` must be one of {self._DRP_EXTRA} or "
                f"{self._RPT_EXTRA} but got {ignored_behavior} instead!"
            )

        if fields_to_unpack is not None:
            field_names = set(fields_to_unpack)
            self.check_unpack_fn = lambda f: f in field_names
        elif fields_to_ignore is not None:
            field_names = set(fields_to_ignore)
            self.check_unpack_fn = lambda f: f not in field_names
        else:
            field_names = None
            # unpack all!
            self.check_unpack_fn = lambda _: True

        self.ignore_behavior = ignored_behavior

        super().__init__(
            input_fields=list(field_names) if field_names else None,
            output_fields=list(field_names) if field_names else None,
        )

    def transform(
        self, data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:

        # we get the names of fields to unpack after seeing the
        # first sample; so we set this to None for now
        all_field_names_to_unpack: Optional[List[str]] = None

        # iterate over all samples
        for packed_sample in data:

            # first, compute the names of fields to unpack
            if all_field_names_to_unpack is None:
                all_field_names_to_unpack = [
                    k for k in packed_sample.keys() if self.check_unpack_fn(k)
                ]
                if len(all_field_names_to_unpack) == 0:
                    # raise an error if there's nothing to unpack,
                    # which might indicate an error in setting up the mapper
                    raise ValueError("No fields to unpack!")

            unpacked_samples_it: Iterable[Dict[str, Any]] = (
                # this re-attached all the field names to just the values
                # of the new unpacked samples.
                {
                    k: v
                    for k, v in zip(
                        all_field_names_to_unpack, unpacked_element_values
                    )
                }
                # this zip goes from list of features, each containing
                # multiple elements in this packed sample, to a list of
                # list of features in a element.
                for unpacked_element_values in zip(
                    *(packed_sample[k] for k in all_field_names_to_unpack)
                )
            )

            if self.ignore_behavior == self._RPT_EXTRA:
                # duplicate the fields that have not been unpacked
                # in all new samples

                features_to_duplicate = {
                    k: v
                    for k, v in packed_sample.items()
                    if k not in all_field_names_to_unpack
                }
                unpacked_samples_it = (
                    # this duplicates the unpacked samples
                    {**unpacked_sample, **features_to_duplicate}
                    for unpacked_sample in unpacked_samples_it
                )

            yield from unpacked_samples_it



import os
import warnings
from typing import Optional, Type


class SmashedWarnings:
    _WARNINGS = bool(os.environ.get("SMASHED_WARNINGS", True))

    @classmethod
    def toggle(cls, value: Optional[bool] = None):
        if value is None:
            value = not cls._WARNINGS
        cls._WARNINGS = value

    @classmethod
    def _warn(
        cls: Type["SmashedWarnings"],
        message: str,
        category: Type[Warning],
        stacklevel: int = 2,
    ):
        if cls._WARNINGS:
            warnings.warn(message, category, stacklevel=stacklevel)

    @classmethod
    def deprecation(cls, message: str):
        cls._warn(message, RuntimeWarning)

    @classmethod
    def precedence(cls, message: str):
        cls._warn(message, RuntimeWarning)



import inspect
from typing import Any, List, Union

from ftfy import TextFixerConfig, fix_text

from ..base.mappers import SingleBaseMapper
from ..base.types import TransformElementType


class FtfyMapper(SingleBaseMapper):
    """Uses ftfy to fix text encoding and general weirdness issues."""

    def __init__(
        self,
        input_fields: Union[str, List[str]],
        **ftfy_kwargs: Any,
    ) -> None:
        """Initialize the mapper.

        Args:
            input_fields (Union[str, List[str]]): The fields to fix;
                if a string, it is assumed to be a single field.
            ftfy_kwargs (Any): Any keyword arguments to use to create a
                ftfy.TextFixerConfig object. See the ftfy documentation
                for more information.
        """

        if isinstance(input_fields, str):
            input_fields = [input_fields]

        self.fields_to_fix = set(input_fields)

        # check if options for ftfy are valid
        valid_ftfy_args = set(inspect.getfullargspec(TextFixerConfig).args)
        for arg_name in ftfy_kwargs:
            if arg_name not in valid_ftfy_args:
                raise ValueError(
                    f"Invalid argument for ftfy.TextFixerConfig: {arg_name}"
                )
        self.ftfy_config = TextFixerConfig(**ftfy_kwargs)

        super().__init__(input_fields=input_fields, output_fields=input_fields)

    def transform(self, data: TransformElementType) -> TransformElementType:
        return {
            field: (
                fix_text(value, config=self.ftfy_config)
                if field in self.fields_to_fix
                else value
            )
            for field, value in data.items()
        }



"""


@lucas, @kylel

"""

import unicodedata
from typing import Any, List, Optional

from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from ..base.mappers import SingleBaseMapper
from ..base.types import TransformElementType


class TokenizerMapper(SingleBaseMapper):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        input_field: str,
        output_prefix: Optional[str] = None,
        add_special_tokens: Optional[bool] = True,
        max_length: Optional[int] = None,
        is_split_into_words: Optional[bool] = False,
        return_token_type_ids: Optional[bool] = False,
        return_attention_mask: Optional[bool] = True,
        return_overflowing_tokens: Optional[bool] = False,
        return_special_tokens_mask: Optional[bool] = False,
        return_offsets_mapping: Optional[bool] = False,
        return_length: Optional[bool] = False,
        return_word_ids: Optional[bool] = False,
        return_words: Optional[bool] = False,
        **tk_kwargs: Any,
    ) -> None:
        self.prefix = f"{output_prefix}_" if output_prefix else ""
        self.to_tokenize_filed = input_field
        self.tokenizer = tokenizer

        # arguments to be passed to the tokenizer __call__ function go here
        tk_kwargs = {
            "add_special_tokens": add_special_tokens,
            "max_length": max_length,
            "is_split_into_words": is_split_into_words,
            **(tk_kwargs or {}),
        }

        output_fields = ["input_ids"]

        # various options for the tokenizer affect which fields are returned
        tk_kwargs["return_attention_mask"] = return_attention_mask
        if return_attention_mask:
            output_fields.append("attention_mask")

        tk_kwargs["return_token_type_ids"] = return_token_type_ids
        if return_token_type_ids:
            output_fields.append("token_type_ids")

        tk_kwargs["return_overflowing_tokens"] = return_overflowing_tokens
        if return_overflowing_tokens:
            output_fields.append("overflow_to_sample_mapping")

        tk_kwargs["return_special_tokens_mask"] = return_special_tokens_mask
        if return_special_tokens_mask:
            output_fields.append("special_tokens_mask")

        tk_kwargs["return_offsets_mapping"] = return_offsets_mapping
        if return_offsets_mapping:
            output_fields.append("offset_mapping")

        tk_kwargs["return_length"] = return_length
        if return_length:
            output_fields.append("length")

        self.return_word_ids = self.return_words = False
        if "is_split_into_words" in tk_kwargs and return_word_ids:
            self.return_word_ids = True
            output_fields.append("word_ids")
            if return_words:
                self.return_words = True
                output_fields.append("words")

        self.tokenize_kwargs = tk_kwargs
        super().__init__(
            input_fields=[self.to_tokenize_filed],
            output_fields=list(map(self._prefixify, output_fields)),
        )

    def _prefixify(self, field_or_dict: str) -> str:
        return f"{self.prefix}{field_or_dict}"

    def transform(self, data: TransformElementType) -> TransformElementType:
        batch_encoding = self.tokenizer(
            data[self.to_tokenize_filed], **self.tokenize_kwargs
        )

        # token_to_word mappings are unfortunately not natively returned by
        # HF.tokenizer; so we need to operate separately on the
        # `batch_encoding` object to get this info.
        # if f'{self.prefix}word_ids' in self.output_fields:
        if self.return_word_ids:
            batch_encoding["word_ids"] = [
                # ignoring pylance complaining because word ids should
                # be provided given that we are passing `is_split_into_words`
                batch_encoding[sequence_id].word_ids  # type: ignore
                for sequence_id in range(
                    # ignoring pylance complaining because input_ids should
                    # always be a list of lists when `is_split_into_words`
                    # is set to True.
                    len(batch_encoding["input_ids"])  # type: ignore
                )
            ]
            if self.return_words:
                batch_encoding["words"] = [
                    [
                        None
                        if word_id is None
                        else data[self.input_fields[0]][word_id]
                        for word_id in word_ids
                    ]
                    # ignoring pylance complaining because we just added
                    # word_ids above!
                    for word_ids in batch_encoding["word_ids"]  # type: ignore
                ]

        return {
            self._prefixify(field_name): field_value
            for field_name, field_value in batch_encoding.items()
        }


class ValidUnicodeMapper(SingleBaseMapper):
    """Given input_fields of type List[str], replaces invalid Unicode
    characters with something else"""

    def __init__(
        self,
        input_fields: List[str],
        unicode_categories: List[str],
        replace_token: str,
        # output_prefix: Optional[str] = None,
    ):
        # self.batched = False
        # self.input_fields = input_fields
        # self.prefix = f'{output_prefix}_' if output_prefix else ''
        # self.output_fields = [
        #     f'{self.prefix}{input_field}'
        #     for input_field in self.input_fields
        # ]
        super().__init__(input_fields=input_fields, output_fields=input_fields)
        self.unicode_categories = unicode_categories
        self.replace_token = replace_token

    def transform(self, data: TransformElementType) -> TransformElementType:
        def _transform(tokens: List[str]) -> List[str]:
            return [
                self.replace_token
                if all(
                    unicodedata.category(ch) in self.unicode_categories
                    for ch in token
                )
                else token
                for token in tokens
            ]

        return {
            k: v if k not in self.input_fields else _transform(v)
            for k, v in data.items()
        }
        # new_data = {f'{self.prefix}{k}': v if k not in self.input_fields
        #             else _transform(v) for k, v in data.items()}
        # return new_data


class PaddingMapper(SingleBaseMapper):
    """Given input_fields of type List[str], figures out how to pad them
    such that all examples in dataset in those fields have same length.
    This can be useful because Huggingface padding has really weird behavior
    for custom keys. For example,

    tokenizer.pad([{'input_ids': [0, 1, 2], 'aaa': [3, 3, 3]},
                   {'input_ids': [3, 4], 'aaa': [4, 4]}])

    will correctly pad `input_ids`, but not `aaa`.  This can break collation
    which often calls `tokenizer.pad`.
    """

    def __init__(
        self,
        pad_to_length: int,
        pad_value: Any,
        fields_to_pad: Optional[List[str]] = None,
    ):
        super().__init__()
        self.pad_to_length = pad_to_length
        self.pad_value = pad_value
        self.fields_to_pad = fields_to_pad

    def transform(self, data: TransformElementType) -> TransformElementType:
        """Add padding to all list elements for the fields we specify."""
        fields_to_pad = (
            data.keys() if self.fields_to_pad is None else self.fields_to_pad
        )

        def _pad(input_elements: List[Any]) -> List[Any]:
            if len(input_elements) > self.pad_to_length:
                raise ValueError(
                    f"PaddingMapper expects every input sequence to be less"
                    f"than or equal to the `pad_to_length`. Please handle"
                    f"any truncation or whatever upstream in a different "
                    f"mapper, such as TokenizerMapper."
                    f"\t{len(input_elements)} > {self.pad_to_length}"
                    f"\t{input_elements}"
                )
            input_elements += [
                self.pad_value
                for _ in range(self.pad_to_length - len(input_elements))
            ]
            return input_elements

        return {
            k: v if k not in fields_to_pad else _pad(v)
            for k, v in data.items()
        }



from typing import TYPE_CHECKING, Any, List, Optional, Sequence, TypeVar

from necessary import necessary
from trouting import trouting

from ..base.mappers import SingleBaseMapper
from ..base.types import TransformElementType

with necessary("datasets", soft=True) as HUGGINGFACE_DATASET_AVAILABLE:
    if HUGGINGFACE_DATASET_AVAILABLE or TYPE_CHECKING:
        from datasets.arrow_dataset import Dataset
        from datasets.iterable_dataset import IterableDataset

        HuggingFaceDataset = TypeVar(
            "HuggingFaceDataset", Dataset, IterableDataset
        )


class ChangeFieldsMapper(SingleBaseMapper):
    def __init__(
        self,
        keep_fields: Optional[List[str]] = None,
        drop_fields: Optional[List[str]] = None,
    ):
        """Mapper that removes some of the fields in a dataset.
        Either `keep_fields` or `drop_fields` must be specified, but not both.

        Args:
            keep_fields (List[str]): Fields to keep, all other fields
                are dropped. Defaults to [].
            drop_fields (List[str]): Fields to drop, all other fields
                are kept. Defaults to [].
        """

        # xor between keep_fields and remove_fields
        if (keep_fields is not None and drop_fields is not None) or (
            keep_fields is None and drop_fields is None
        ):
            raise ValueError("Must specify `keep_fields` or `drop_fields`")

        super().__init__(input_fields=drop_fields, output_fields=keep_fields)

    @trouting
    def map(  # type: ignore
        self,
        dataset: Any,
        **map_kwargs: Any,
    ) -> Any:
        return super().map(dataset, **map_kwargs)

    @map.add_interface(dataset=list)
    def map_list_of_dicts(
        self,
        dataset: Sequence[TransformElementType],
        **map_kwargs: Any,
    ) -> Sequence[TransformElementType]:
        """If the dataset is a list of dicts, we need to make sure that
        all existing columns are removed."""
        return super().map(dataset, remove_columns=True, **map_kwargs)

    if HUGGINGFACE_DATASET_AVAILABLE:

        @map.add_interface(dataset=(Dataset, IterableDataset))
        def map_huggingface_dataset(
            self,
            dataset: HuggingFaceDataset,
            **map_kwargs: Any,
        ) -> HuggingFaceDataset:
            """If the dataset is a HuggingFace dataset, we also need to
            make sure that existing columns get removed, but in this case
            the mechanism is different."""

            # mechanism to remove columns in huggingface datasets is
            # slightly different than in list of dict datasets.
            map_kwargs = {
                "remove_columns": list(dataset.features.keys()),
                **map_kwargs,
            }
            return super().map(dataset, **map_kwargs)

    def transform(self, data: TransformElementType) -> TransformElementType:
        if self.input_fields:
            new_data = {
                k: v for k, v in data.items() if k not in self.input_fields
            }

        elif self.output_fields:
            new_data = {k: data[k] for k in self.output_fields}

        else:
            raise ValueError("Must specify `keep_fields` or `drop_fields`")

        return new_data


class MakeFieldMapper(SingleBaseMapper):
    def __init__(
        self: "MakeFieldMapper",
        field_name: str,
        value: Any,
        shape_like: Optional[str] = None,
    ):
        """Mapper that adds a new field to a dataset.

        Args:
            field_name (str): Name of the new field.
            value (Optional[Any]): Value to assign to the new field.
            shape_like (Optional[str], optional): If a fixed value is provided,
                this existing field that will be used to determine the shape of
                the new field. Defaults to None.

        """
        super().__init__(output_fields=[field_name])
        self.value = value
        self.shape_like = shape_like

    def transform(self, data: TransformElementType) -> TransformElementType:
        if self.shape_like is not None:
            new_value = [self.value for _ in data[self.shape_like]]
        else:
            new_value = self.value

        data[self.output_fields[0]] = new_value
        return data



import functools
from collections import abc
from itertools import chain
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import torch
from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from ..base.mappers import SingleBaseMapper
from ..base.types import TransformElementType

__all__ = [
    "ListCollatorMapper",
    "TensorCollatorMapper",
    "FromTokenizerListCollatorMapper",
    "FromTokenizerTensorCollatorMapper",
]


class CollatorMixIn:
    def __init__(
        self,
        pad_to_length: Optional[Union[int, Sequence[int]]] = None,
        pad_to_length_dim: Optional[int] = None,
        fields_pad_ids: Optional[Mapping[str, int]] = None,
        unk_fields_pad_id: Optional[int] = None,
    ):
        """Create a collator.

        Args:
            pad_to_length (Union[int, Sequence[int]], optional): If provided
                and is single value, pad all sequences to this length. If
                provided  and is sequence, we assume we should pad each
                dimension to the  corresponding length. In both cases, we will
                raise an error if any sequence is longer than the requested
                length. When not provided or None, sequences will be padded to
                the length of the longest sequence. Defaults to None.
            fields_pad_ids (Mapping[str, int], optional): A mapping from field
                names to the padding value to use for that field. If not
                provided, the mapper will fail unless the unk_fields_pad_id
                attribute is set.
            unk_fields_pad_id (int, optional): The padding value to use for
                any field that is not in fields_pad_ids. If not provided, an
                error will be raised if a field is not in fields_pad_ids.
        """
        self.fields_pad_ids = fields_pad_ids or {}
        self.pad_to_length = pad_to_length
        self.unk_fields_pad_id = unk_fields_pad_id

        if self.unk_fields_pad_id is None and self.fields_pad_ids is None:
            raise ValueError(
                "Either `unk_fields_pad_id` or `fields_to_pad` "
                "must be provided"
            )

        super().__init__()

    def _get_padding_value(self, field_name: str) -> int:
        if field_name in self.fields_pad_ids:
            return self.fields_pad_ids[field_name]
        elif self.unk_fields_pad_id is not None:
            return self.unk_fields_pad_id
        else:
            raise ValueError(
                f"Must specify a padding value for field {field_name}"
                f"or provide a extra_fields_padding_id attribute to "
                "the mapper to handle unrecognized fields"
            )


class FromTokenizerMixIn(CollatorMixIn):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        pad_to_length: Optional[Union[int, Sequence[int]]] = None,
        fields_pad_ids: Optional[Mapping[str, int]] = None,
        unk_fields_pad_id: Optional[int] = None,
    ):
        """Create a collator for tensors.

        Args:
            tokenizer (PreTrainedTokenizerBase): The tokenizer to use to pad
                common fields.
            pad_to_length (Union[int, Sequence[int]], optional): If provided
                and is single value, pad all sequences to this length. If
                provided  and is sequence, we assume we should pad each
                dimension to the  corresponding length. In both cases, we will
                raise an error if any sequence is longer than the requested
                length. When not provided or None, sequences will be padded to
                the length of the longest sequence. Defaults to None.
            fields_pad_ids (Mapping[str, int], optional): A mapping from field
                names to the padding value to use for that field. If not
                provided, the mapper will fail unless the unk_fields_pad_id
                attribute is set.
            unk_fields_pad_id (int, optional): The padding value to use for
                any field that is not in fields_pad_ids. If not provided, an
                error will be raised if a field is not in fields_pad_ids.
        """

        fields_pad_ids = {
            "input_ids": tokenizer.pad_token_id or 0,
            "attention_mask": 0,
            "token_type_ids": tokenizer.pad_token_type_id or 0,
            "overflow_to_sample_mapping": 0,
            "special_tokens_mask": 0,
            "offset_mapping": 0,
            "length": 0,
            **(fields_pad_ids or {}),
        }
        super().__init__(
            pad_to_length=pad_to_length,
            unk_fields_pad_id=unk_fields_pad_id,
            fields_pad_ids=fields_pad_ids,
        )


class TensorCollatorMapper(CollatorMixIn, SingleBaseMapper):
    """
    A collator mapper that collates sequences of n tensors into a single
    tensor of shape (n, ...) where ... is the maximum size of the
    sequences. Uses the values passed to fields_pad_ids to determine
    how to pad each field.

    If used with a Pytorch DataLoader, the collator mapper must be
    called as follows:

    >>> collator = TensorsCollatorMapper(...)
    >>> data_loader = DataLoader(..., collate_fn=collator.transform)
    """

    @staticmethod
    def _pad(
        sequence: Sequence[torch.Tensor],
        pad_value: int,
        dim: int = 0,
        pad_to_length: Optional[Union[int, Sequence[int]]] = None,
    ) -> torch.Tensor:

        # make sure type of input is right
        if not (
            isinstance(sequence, abc.Sequence)
            and all(isinstance(elem, torch.Tensor) for elem in sequence)
        ):
            raise ValueError(
                "Each element to collate must be a sequence of torch.Tensor, "
                f"not {type(sequence)}"
            )

        # the `view` is because we need to add a new dimension as the
        # dimension alongside which we batch
        sequence = [torch.unsqueeze(tensor, dim=dim) for tensor in sequence]

        # this contains maximum length of all the sequences
        max_lengths = [max(t) for t in zip(*(t.size() for t in sequence))]

        if isinstance(pad_to_length, int):
            # if pad_to_length is a single integer, we pad all sequences to
            # the same length
            pad_to_length = [pad_to_length] * len(sequence)

        # if pad_to_length is provided, we need to make sure that
        # each dimension does not exceed the pad_to_length[i].
        if pad_to_length is not None:
            for i in range(len(max_lengths)):
                if i == dim:
                    # do not touch the dimension we are batching along
                    continue
                elif max_lengths[i] > pad_to_length[i]:
                    # raise an error if the pad_to_length requested
                    # for this dimension is smaller than the maximum
                    # length of the current sequence of this dimension
                    raise ValueError(
                        f"Tried to pad sequence to length {pad_to_length} "
                        f"but sequence had shape {max_lengths}"
                    )
                else:
                    max_lengths[i] = pad_to_length[i]

        # https://pytorch.org/docs/stable/generated/torch.nn.functional.pad
        # according to that page, we need to create pad shapes is reverse.
        pad_shapes = tuple(
            tuple(
                chain.from_iterable(
                    (0, m - s)
                    for s, m in zip(t.size()[::-1], max_lengths[::-1])
                )
            )
            # we do padding shapes for each tensor
            for t in sequence
        )
        # call each pad on each of the tensors with the appropriate padding
        to_stack = tuple(
            torch.nn.functional.pad(
                tensor, pad, mode="constant", value=pad_value
            )
            for tensor, pad in zip(sequence, pad_shapes)
        )

        return torch.cat(to_stack, dim=dim)

    def transform(
        self: "TensorCollatorMapper", data: Dict[str, Sequence[torch.Tensor]]
    ) -> Dict[str, torch.Tensor]:

        collated_data = {
            field_name: self._pad(
                sequence=list_of_tensors,
                pad_value=self._get_padding_value(field_name=field_name),
                pad_to_length=self.pad_to_length,
            )
            for field_name, list_of_tensors in data.items()
        }
        return collated_data


class FromTokenizerTensorCollatorMapper(
    FromTokenizerMixIn, TensorCollatorMapper
):
    """
    A collator mapper that collates sequences of n tensors into a
    single tensor of shape (n, ...) where ... is the maximum size of the
    sequences. Uses the provided tokenizer to determine how to pad common
    fields for NLP tasks, such as `input_ids`, `attention_mask`,
    `token_type_ids`, etc. Padding values for further fields can be
    provided in the `fields_pad_ids` argument.

    If used with a Pytorch DataLoader, the collator mapper must be
    called as follows:
    >>> tokenizer = AutoTokenizer.from_pretrained(...)
    >>> collator = FromTokenizerTensorsCollatorMapper(tokenizer, ...)
    >>> data_loader = DataLoader(..., collate_fn=collator.transform)
    """


class ListCollatorMapper(CollatorMixIn, SingleBaseMapper):
    """Given a `sequence_to_pad` dict of comprised of <field names, value to
    pad field with>, figures out how much to pad them such that all examples in
    dataset in those fields have same length. This mapper can be useful for
    cases where HuggingFace tokenizer padding behaves weirdly, which is the
    case when it is customized to produce extra keys. For example,

    tokenizer.pad([{'input_ids': [0, 1, 2], 'aaa': [3, 3, 3]},
                   {'input_ids': [3, 4], 'aaa': [4, 4]}])

    will correctly pad `input_ids`, but not `aaa`.  This can break collation
    because it requires calls to `tokenizer.pad`.
    """

    pad_to_length: int

    def _get_list_shape_recursive(
        self, sequence: Sequence[Any]
    ) -> Tuple[int, ...]:
        if not isinstance(sequence, abc.Sequence):
            # we have unpacked as far as we can go; this is just a
            # single element.
            return tuple()

        # this iterator will yield the shape of each element in the sequence
        inner_dims = (self._get_list_shape_recursive(s) for s in sequence)

        # the acutal shape is the maximum of the inner dims
        inner_shape = tuple(max(dims) for dims in zip(*inner_dims))

        return (len(sequence), *inner_shape)

    def _pad_recursive(
        self, sequence: List[Any], shape: Sequence[int], padding_symbol: Any
    ) -> List[Any]:
        """Recursively pads a list of [lists, ...].

        Args:
            sequence (List[Any]): The list to pad.
            shape (Sequence[int]): The shape to pad to.
            padding_symbol (Any): The symbol to pad with.

        Returns:
            List[Any]: _description_
        """

        if len(shape) < 2:
            # we have reached the end of the shape; this is a single element
            return sequence

        _, dim_to_pad_shape, *rest_shape = shape

        nested_pad_symbol = functools.reduce(
            lambda x, _: [x], range(len(rest_shape)), padding_symbol
        )

        # Let's walk through the nested padding process here.
        #
        # Our overall goal is to turn a potentially nested sequence of
        # list such as:
        #
        #       [[[1 2 3] [4 5]    ]
        #        [[6 7]   [8]   [9]]]
        #
        # Into a list of lists of the same length, such as:
        #
        #       [[[1 2 3] [4 5 0] [0 0 0]]
        #        [[6 7 0] [8 0 0] [9 0 0]]]
        #
        # The first step is to add any sequence that is completely missing
        # such as the last example of the first row above. That will get us
        # the following:
        #
        #      [[[1 2 3] [4 5] [0]]
        #       [[6 7]   [8]   [9]]]
        #
        # We do that in the following line:
        sequence_with_brand_new_padding = (
            sub_seq + [nested_pad_symbol] * (dim_to_pad_shape - len(sub_seq))
            for sub_seq in sequence
        )

        # The second step is to recursively pad the inner lists by
        # calling this function on each subsequence. We do that as follows:
        padded_sequence = [
            self._pad_recursive(
                sequence=sub_seq,
                shape=(dim_to_pad_shape, *rest_shape),
                padding_symbol=padding_symbol,
            )
            for sub_seq in sequence_with_brand_new_padding
        ]

        return padded_sequence

    def _pad(
        self: "ListCollatorMapper",
        seq_of_seq_to_pad: List[Any],
        padding_symbol: Any,
    ) -> List[Any]:

        padding_shape = self._get_list_shape_recursive(seq_of_seq_to_pad)

        if self.pad_to_length is not None:
            if not all(p <= self.pad_to_length for p in padding_shape):
                raise ValueError(
                    "PaddingMapper expects every input sequence to be less"
                    "than or equal to the `pad_to_length`. Please handle"
                    "any truncation or whatever upstream in a different"
                    " mapper, such as TokenizerMapper."
                    f"\t{padding_shape} > {self.pad_to_length}"
                    f"\t{seq_of_seq_to_pad}"
                )
            padding_shape = (self.pad_to_length,) * len(padding_shape)

        if len(padding_shape) < 2:
            # nothing to pad here; we need at minimum a list of lists
            # for padding to make any sense.
            return seq_of_seq_to_pad

        padded_sequence = self._pad_recursive(
            sequence=seq_of_seq_to_pad,
            shape=padding_shape,
            padding_symbol=padding_symbol,
        )
        return padded_sequence

    def transform(self, data: TransformElementType) -> TransformElementType:
        """Add padding to all list elements for the fields we specify."""

        return {
            field_name: self._pad(
                seq_of_seq_to_pad=field_value,
                padding_symbol=self._get_padding_value(field_name=field_name),
            )
            for field_name, field_value in data.items()
        }


class FromTokenizerListCollatorMapper(FromTokenizerMixIn, ListCollatorMapper):
    """Performs collation of a list to a given length. Uses the provided
    tokenizer to determine how to pad common fields for NLP tasks, such as
    `input_ids`, `attention_mask`, `token_type_ids`, etc. Padding values for
    further fields can be provided in the `fields_pad_ids` argument."""



from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, TypeVar, Union

import torch
from necessary import necessary
from trouting import trouting

from ..base.mappers import SingleBaseMapper
from ..base.types import TransformElementType

with necessary("datasets", soft=True) as HUGGINGFACE_DATASET_AVAILABLE:
    if HUGGINGFACE_DATASET_AVAILABLE or TYPE_CHECKING:
        from datasets.arrow_dataset import Dataset
        from datasets.iterable_dataset import IterableDataset

        HuggingFaceDataset = TypeVar(
            "HuggingFaceDataset", Dataset, IterableDataset
        )


class Python2TorchMapper(SingleBaseMapper):

    __slots__ = ["field_cast_map", "device"]
    field_cast_map: Dict[str, torch.dtype]
    device: Union[torch.device, None]

    def __init__(
        self: "Python2TorchMapper",
        field_cast_map: Optional[Mapping[str, Union[str, torch.dtype]]] = None,
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """Mapper that converts Python types to Torch types. It can optionally
        cast the values of a field to a specific type, and move to a specific
        device.

        Args:
            field_cast_map (Mapping[str, Union[str, torch.dtype]], optional):
                Mapping from field names to the types to cast the values to.
                Defaults to None, which means no casting occurs.
            device (Union[torch.device, str], optional): Device to move the
                tensors to. Defaults to None, which means no moving occurs.
        """
        self.device = torch.device(device) if device else None

        self.field_cast_map = {
            field_name: self._get_dtype(field_type)
            for field_name, field_type in (field_cast_map or {}).items()
        }
        super().__init__(
            input_fields=list(self.field_cast_map.keys()),
            output_fields=list(self.field_cast_map.keys()),
        )

    @staticmethod
    def _get_dtype(dtype: Any) -> torch.dtype:

        if isinstance(dtype, str):
            dtype = getattr(torch, dtype, None)
            if dtype is None:
                raise ValueError(f"Unknown dtype {dtype}")

        if not isinstance(dtype, torch.dtype):
            raise ValueError(f"{dtype} is not a torch dtype")

        return dtype

    def transform(self, data: TransformElementType) -> TransformElementType:
        return {
            # .to(type) and .to(device) will return the original tensor if
            # the type and device are the same/None; this is because
            # for any torch.Tensor t, hash(t) == hash(t.to(None).to(None))
            field_name: torch.tensor(field_value)
            .to(self.field_cast_map.get(field_name, None))
            .to(self.device)
            for field_name, field_value in data.items()
        }

    @trouting
    def map(  # type: ignore
        self,
        dataset: Any,
        **map_kwargs: Any,
    ) -> Any:
        # we need this map to be able to add the new interface below
        # and handle types for which we don't have a new interface but our
        # parent class has one
        return super().map(dataset, **map_kwargs)

    if HUGGINGFACE_DATASET_AVAILABLE:

        @map.add_interface(dataset=(Dataset, IterableDataset))
        def map_huggingface_dataset(
            self,
            dataset: HuggingFaceDataset,
            **map_kwargs: Any,
        ) -> HuggingFaceDataset:
            return dataset.with_format("torch")


class Torch2PythonMapper(SingleBaseMapper):
    def __init__(self: "Torch2PythonMapper") -> None:
        """Mapper that converts Torch types to Python types. It relies on
        function `.tolist()` to convert the tensor to a list. It additionally
        moves the tensor to the CPU before converting it.
        """
        super().__init__()

    def transform(
        self: "Torch2PythonMapper", data: Dict[str, torch.Tensor]
    ) -> TransformElementType:
        return {
            field_name: field_value.cpu().tolist()
            for field_name, field_value in data.items()
        }

    @trouting
    def map(  # type: ignore
        self,
        dataset: Any,
        **map_kwargs: Any,
    ) -> Any:
        # we need this map to be able to add the new interface below
        # and handle types for which we don't have a new interface but our
        # parent class has one
        return super().map(dataset, **map_kwargs)

    if HUGGINGFACE_DATASET_AVAILABLE:

        @map.add_interface(dataset=(Dataset, IterableDataset))
        def map_huggingface_dataset(
            self,
            dataset: HuggingFaceDataset,
            **map_kwargs: Any,
        ) -> HuggingFaceDataset:
            return dataset.with_format(None)



from typing import Iterable

from ..base.mappers import BatchedBaseMapper
from ..base.types import TransformElementType


class FixedBatchSizeMapper(BatchedBaseMapper):
    def __init__(
        self: "FixedBatchSizeMapper",
        batch_size: int,
        keep_last: bool = True,
    ) -> None:
        """A very simple mapper that batches samples into fixed-size batches.

        Args:
            batch_size (int): The number of samples in each batch.
            keep_last (bool, optional): Whether to keep the last batch if it
                is not full. Defaults to True.
        """
        try:
            batch_size = int(batch_size)
            assert batch_size > 0
        except Exception:
            raise ValueError(
                f"batch_size must be a positive integer, not {batch_size}"
            )

        if not isinstance(keep_last, bool):
            raise ValueError(f"keep_last must be a boolean, not {keep_last}")

        self.batch_size = batch_size
        self.keep_last = keep_last
        super().__init__()

    def transform(
        self: "FixedBatchSizeMapper", data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:
        accumulator = None
        counter = 0
        for sample in data:
            if accumulator is None:
                accumulator = {k: [v] for k, v in sample.items()}
            else:
                [accumulator[k].append(v) for k, v in sample.items()]
            counter += 1

            if counter == self.batch_size:
                yield accumulator
                accumulator = None
                counter = 0

        if self.keep_last and accumulator is not None:
            yield accumulator
